//+-----------------------------------------------------------------------------+
//| Copyleft (C) 2016 REGEO (http://www.regeo.uji.es/)
//|
//| This program is free software
//| You can redistribute it and/or modify it under the terms of the GNU General Public License,
//| as published by the Free Software Foundation.
//| 
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//+----------------------------------------------------------------------------*/

#include "stdafx.h"
#include <fstream>
#include <cmath>
#include <cassert>
#include <vector>

#include <iostream>
#include <string>
#include <fstream>

#include "Tools_Vector.h"
#include "Tools_Geometry.h"

#include "CueJunctions.h"


CModelJunctions::CModelJunctions()
//Model for Junctions detection
{
	VertexCount = 0;
	Junction.clear();

	EdgeCount = 0;
	EdgeHead.clear();
	EdgeTail.clear();
}


CModelJunctions::CModelJunctions(const CModelJunctions& MJ)
	:VertexCount(MJ.VertexCount),
	Junction(MJ.Junction),
	EdgeCount(MJ.EdgeCount),
	EdgeHead(MJ.EdgeHead),
	EdgeTail(MJ.EdgeTail)
{
}


CModelJunctions::~CModelJunctions()
{
}

void CModelJunctions::SetPOINT2D(long Position, POINT2D const &POINT2D)
{
	assert(Position>=0 && Position<(long)Junction.size());
	Junction[Position]= POINT2D;
}

double CModelJunctions::GetLength2D(long IndexEdge)
//Gets the length of the edge "IndexEdge"
//Considers only X and Y coordinates
{
	POINT2D A= Junction[EdgeHead[IndexEdge]];
	POINT2D B= Junction[EdgeTail[IndexEdge]];
	return MODULE_2D(A, B);
}

void CModelJunctions::RemovePOINT2D(long Position, bool updateEdges)
//Removes the vertex "Position"

//If the flag "update" is true, edges connected to
//vertices subsequent to Position are renumbered
//(This is de default, to prevent inconsistencies in the database)
{
	RemoveP(Position);

	if (updateEdges){
		//Vertices of edges connected to vertices subsequent to Position are renumbered
		for (long k=0; k< (long) EdgeCount; k++){
			if (EdgeTail[k] > Position){
				EdgeTail[k]--;
			}
			if(EdgeHead[k] > Position){
				EdgeHead[k]--;
			}
		}
	}
}

void CModelJunctions::RemoveP(long Position)
{
	assert(Position>=0 && Position<(long)VertexCount);
	Junction.erase(Junction.begin()+Position);
	VertexCount--;
}

std::vector<long> CModelJunctions::GetAllEdgesInVertex(long iVertex)
// Returns the edges connected to the vertex "iVertex"
{
	std::vector<long> Destination;
	for(long i= 0; i<(long) EdgeCount; i++){
		if(EdgeHead[i]==iVertex || EdgeTail[i]==iVertex)
			Destination.push_back(i);
	}
	return Destination;
}

long CModelJunctions::GetSizeEdgesInVertex(long iPoint)
// Returns the total number of edges connected to the vertex "iVertex"
{
	long NumEdges= 0;
	for(long i= 0; i<(long)EdgeCount; i++){
		if(EdgeHead[i]==iPoint || EdgeTail[i]==iPoint)
			NumEdges++;
	}
	return NumEdges;
}

//---------------------------------------------------
///////////////////////////////////////////////////////////////
//Delaunay triangulation adapted from:
//
// Gilles Dumoulin's C++ implementation of Paul Bourke's Delaunay triangulation algorithm.
// Source: http://paulbourke.net/papers/triangulate/


bool Junctions2D::Delaunay::ReorderPoints(std::vector<POINT2D> &p,
										  std::vector<long> &NumPoint)
//Bubble sort points from smaller to larger x coordinate

//The list of renumbered points is also returned
{
	long nv= p.size();

	for (long i=0; i<nv; i++){
		NumPoint.push_back(i);
	}

	bool swapped;
	do{
		swapped= false;
		for (long i= 0; i<nv-1; i++){
			if (p[i].x > p[i+1].x){
				std::swap(NumPoint[i], NumPoint[i+1]);
				std::swap(p[i], p[i+1]);
				swapped= true;
			}
		}
	}while(swapped);

	//Check for repeated points
	for (long i=0; i<nv-1; i++){
		if (p[i].x == p[i+1].x && p[i].y == p[i+1].y){
			//Triangulation fails for clouds with repeated points!!
			return false;
		}
	}
	return true;

}


bool Junctions2D::Delaunay::CircumCircle(double xp, double yp, double x1, double y1,
										 double x2, double y2, double x3, double y3,
										 double &xc, double &yc, double &r)
// Return true if a point (xp,yp) is inside the circumcircle made up
// of the points (x1,y1), (x2,y2), (x3,y3)
// Note : A point on the edge is inside the circumcircle

// Also returns the circumcircle centre in (xc,yc), and the radius r
{
	double EPSILON = 0.000001;
	double m1, m2, mx1, mx2, my1, my2;

	// Check for coincident points
	if(abs(y1 - y2) < EPSILON && abs(y2 - y3) < EPSILON)
		return(false);

	if(abs(y2-y1) < EPSILON){ 
		m2 = - (x3 - x2) / (y3 - y2);
		mx2 = (x2 + x3) / 2.0;
		my2 = (y2 + y3) / 2.0;

		xc = (x2 + x1) / 2.0;
		yc = m2 * (xc - mx2) + my2;
	}
	else if(abs(y3 - y2) < EPSILON){ 
		m1 = - (x2 - x1) / (y2 - y1);
		mx1 = (x1 + x2) / 2.0;
		my1 = (y1 + y2) / 2.0;

		xc = (x3 + x2) / 2.0;
		yc = m1 * (xc - mx1) + my1;
	}
	else{
		m1 = - (x2 - x1) / (y2 - y1); 
		m2 = - (x3 - x2) / (y3 - y2); 
		mx1 = (x1 + x2) / 2.0; 
		mx2 = (x2 + x3) / 2.0;
		my1 = (y1 + y2) / 2.0;
		my2 = (y2 + y3) / 2.0;

		xc = (m1 * mx1 - m2 * mx2 + my2 - my1) / (m1 - m2); 
		yc = m1 * (xc - mx1) + my1; 
	}

	double dx = x2 - xc;
	double dy = y2 - yc;
	double rsqr = dx * dx + dy * dy;
	r = sqrt(rsqr); 
	dx = xp - xc;
	dy = yp - yc;
	double drsqr = dx * dx + dy * dy;

	return((drsqr <= rsqr) ? true : false);
}


void Junctions2D::Delaunay::Triangulate(std::vector<POINT2D> point,
										std::vector<TRIANGLE> &v,
										int &ntri)
//Triangulation subroutine
//Takes as input NV vertices in array point
//Returned is a list of ntri triangular faces in the array v
//(These triangles are arranged in a consistent clockwise order)

//The vertex array must be sorted in increasing x values
{
	int nv= point.size();

	//The vector point must be big enough to hold 3 more points
	//(used to save the vertices of the supertriangle)
	POINT2D p;
	p.x= 0;
	p.y= 0;
	point.push_back(p);
	point.push_back(p);
	point.push_back(p);

	//Allocate memory for the completeness list, flag for each triangle
	std::vector<bool> complete;

	//Find the maximum and minimum vertex bounds.
	//This is to allow calculation of the bounding triangle
	double xmin = point[0].x;
	double ymin = point[0].y;
	double xmax = xmin;
	double ymax = ymin;
	for(long i = 1; i < nv; i++){
		if (point[i].x < xmin) xmin = point[i].x;
		if (point[i].x > xmax) xmax = point[i].x;
		if (point[i].y < ymin) ymin = point[i].y;
		if (point[i].y > ymax) ymax = point[i].y;
	}
	double dx = xmax - xmin;
	double dy = ymax - ymin;
	double dmax = (dx > dy) ? dx : dy;
	double xmid = (xmax + xmin) / 2.0;
	double ymid = (ymax + ymin) / 2.0;

	//Set up the supertriangle that encompasses all the sample points
	//(The supertriangle coordinates are added to the end of the vertex list)
	//(The supertriangle is the first triangle in the triangle list)
	point[nv+0].x = xmid - 20 * dmax;
	point[nv+0].y = ymid - dmax;
	point[nv+1].x = xmid;
	point[nv+1].y = ymid + 20 * dmax;
	point[nv+2].x = xmid + 20 * dmax;
	point[nv+2].y = ymid - dmax;

	TRIANGLE t;
	t.p1= nv;
	t.p2= nv+1;
	t.p3= nv+2;
	v.push_back(t);

	complete.push_back(false);
	ntri = 1;

	//Include each point, one at a time, into the existing mesh
	for(long i=0; i<nv; i++){
		double xp = point[i].x;
		double yp = point[i].y;

		//Allocate memory for the edge list
		int nedge = 0;
		std::vector<EDGE> edges;

		//Set up the edge buffer
		//(If the point (xp,yp) lies inside the circumcircle then the
		//three edges of that triangle are added to the edge buffer
		//and that triangle is removed)
		for(long j=0; j<ntri; j++){
			if(complete[j])
				continue;

			double xc, yc, r;
			bool inside = CircumCircle(xp, yp,
									   point[v[j].p1].x, point[v[j].p1].y,
									   point[v[j].p2].x, point[v[j].p2].y,
									   point[v[j].p3].x, point[v[j].p3].y,
									   xc, yc, r);

			if (xc < xp && ((xp-xc)*(xp-xc)) > r)
				complete[j] = true;

			if(inside){
				EDGE e;
				e.p1= v[j].p1; e.p2= v[j].p2;
				edges.push_back(e);
				e.p1= v[j].p2; e.p2= v[j].p3;
				edges.push_back(e);
				e.p1= v[j].p3; e.p2= v[j].p1;
				edges.push_back(e);
				nedge += 3;

				v[j] = v[ntri-1];
				v.erase(v.begin()+v.size()-1);
				complete[j] = complete[ntri-1];
				complete.erase(complete.begin()+complete.size()-1);
				ntri--;

				j--;
			}
		}

		//Tag multiple edges
		//(Note: if all triangles are specified anticlockwise
		//then all interior edges are opposite pointing in direction)
		for(long j=0; j<nedge-1; j++){
			for(long k=j+1; k<nedge; k++){
				if((edges[j].p1 == edges[k].p2) && (edges[j].p2 == edges[k].p1)){
					edges[j].p1 = -1;
					edges[j].p2 = -1;
					edges[k].p1 = -1;
					edges[k].p2 = -1;
				}
				//Shouldn't need the following, see note above
				if((edges[j].p1 == edges[k].p1) && (edges[j].p2 == edges[k].p2)){
					edges[j].p1 = -1;
					edges[j].p2 = -1;
					edges[k].p1 = -1;
					edges[k].p2 = -1;
				}
			}
		}

		//Form new triangles for the current point
		//Skipping over any tagged edges
		//All edges are arranged in clockwise order
		for(long j=0; j<nedge; j++) {
			if(edges[j].p1 < 0 || edges[j].p2 < 0)
				continue;

			t.p1= edges[j].p1;
			t.p2= edges[j].p2;
			t.p3= i;
			v.push_back(t);

			complete.push_back(false);
			ntri++;
		}
	}

	//Remove triangles with supertriangle vertices
	//These are triangles which have a vertex number greater than nv
	for(long i=0; i<ntri; i++) {
		if(v[i].p1 >= nv || v[i].p2 >= nv || v[i].p3 >= nv){
			v[i] = v[ntri-1];
			v.erase(v.begin()+v.size()-1);
			ntri--;
			i--;
		}
	}

	return;
}


void Junctions2D::Delaunay::addNewTip(long n, std::vector<long> &v)
//Add the number n into the vector v,
//if number n is not yet included in the vector

//The numbers are ordered from smaller to larger
{
	for(long i= 0; i<(long)v.size(); i++){
		if(n == v[i])
			return;  //Does not add repeated numbers
		else if(n < v[i]){
			v.insert(v.begin()+i, n);
			return;
		}
	}
	v.push_back(n);
}


bool Junctions2D::Delaunay::NeighborTips(CModelJunctions MJ_main,
										 std::vector<std::vector<long>> &nt)
//Given a set of tips of lines, its Delaunay triangulation is calculated

//Then, all the tips sharing one triangle with a given tip
//are saved as its Neighbors
//But tips that are connected to each other by a line are not elegible as neighbors
{
	//Get the cloud of tips to triangulate
	//------------------------------------
	std::vector<POINT2D> tips;
	POINT2D point;

	for (long i=0; i<(long) MJ_main.VertexCount; i++){
		POINT2D vi= MJ_main.Junction[i];
		point.assign(vi.x, vi.y);
		tips.push_back(point);
	}

	//Sort the cloud of tips in increasing x values
	//(and save the original numbers of the reordered tips)
	//-----------------------------------------------------
	std::vector<long> NumTip;
	if(!ReorderPoints(tips, NumTip))
		return false;

	//Get the Delaunay triangulation of the tips
	//------------------------------------------
	std::vector<TRIANGLE> v;
	int ntri = 0;
	Triangulate(tips, v, ntri);

	//Get the Neighbors of the tips
	//(while keeping their original numbers)
	//--------------------------------------
	std::vector<std::vector<long>> noNeighbor;
	for (long i=0; i<(long)MJ_main.VertexCount; i++){
		nt.push_back(std::vector<long>());
		noNeighbor.push_back(std::vector<long>());
	}

	//(Tips already connected to each other are not elegible as neighbors)
	for (long i=0; i<(long)MJ_main.EdgeCount; i++){
		long Head= MJ_main.EdgeHead[i];
		long Tail= MJ_main.EdgeTail[i];
		noNeighbor[Head].push_back(Tail);
		noNeighbor[Tail].push_back(Head);
	}

	//Add neighbors for each vertex
	for (long i=0; i<(long)MJ_main.VertexCount; i++){
		for (long j=0; j<ntri; j++){
			//Add the other two corners of those triangles that contain the current vertex
			if(NumTip[v[j].p1] == NumTip[i]){
				if(!IsValueInVector(NumTip[v[j].p2], noNeighbor[NumTip[i]]))
					addNewTip(NumTip[v[j].p2], nt[NumTip[i]]);
				if(!IsValueInVector(NumTip[v[j].p3], noNeighbor[NumTip[i]]))
					addNewTip(NumTip[v[j].p3], nt[NumTip[i]]);
			}
			else if(NumTip[v[j].p2] == NumTip[i]){
				if(!IsValueInVector(NumTip[v[j].p1], noNeighbor[NumTip[i]]))
					addNewTip(NumTip[v[j].p1], nt[NumTip[i]]);
				if(!IsValueInVector(NumTip[v[j].p3], noNeighbor[NumTip[i]]))
					addNewTip(NumTip[v[j].p3], nt[NumTip[i]]);
			}
			else if(NumTip[v[j].p3] == NumTip[i]){
				if(!IsValueInVector(NumTip[v[j].p1], noNeighbor[NumTip[i]]))
					addNewTip(NumTip[v[j].p1], nt[NumTip[i]]);
				if(!IsValueInVector(NumTip[v[j].p2], noNeighbor[NumTip[i]]))
					addNewTip(NumTip[v[j].p2], nt[NumTip[i]]);
			}
		}
	}

	return true;
}

//------------------------------------------------
Junctions2D::Junctions2D(long i_MergeCarefulLineTips,
						 double i_maxDistance,
						 double i_maxRotation,
						 double i_ReificationMerit,
						 long i_MergeCasualLineTips,
						 double i_CasualInTol,
						 double i_CasualOutTol,
						 long i_ReMergeTriplets,
						 double i_maxDistanceTriplets,
						 double i_EmergentTripletsMerit,
						 long i_ReMergeDangling,
						 double i_maxDistanceDangling,
						 double i_EmergenceDanglingDeMerit)
//Initialize with input parameters
: MergeCarefulLineTips(i_MergeCarefulLineTips),
  maxDistance(i_maxDistance),
  maxRotation (i_maxRotation),
  ReificationMerit (i_ReificationMerit),
  MergeCasualLineTips(i_MergeCasualLineTips),
  CasualInTol(i_CasualInTol),
  CasualOutTol(i_CasualOutTol),
  ReMergeTriplets (i_ReMergeTriplets),
  maxDistanceTriplets (i_maxDistanceTriplets),
  EmergentTripletsMerit(i_EmergentTripletsMerit),
  ReMergeDangling (i_ReMergeDangling),
  maxDistanceDangling (i_maxDistanceDangling),
  EmergenceDanglingDeMerit (i_EmergenceDanglingDeMerit)
{
}

Junctions2D::~Junctions2D()
{
}

void Junctions2D::Input()
//Copies information from input file
//into the local database

//The intput required by the algorithms is:
//		TipCount= number of Tips in the wireframe
//		TipX[i]= X coordinate of the i-th tip
//		TipY[i]= Y coordinate of the i-th tip

//		EdgeCount= number of edges in the wireframe
//		EdgeHead[i]= Head tip defining the i-th edge
//		EdgeTail[i]= Tail tip defining the i-th edge

//This function should be adapted to different external databases
{
	//Clear data
	//----------
	MJ_main.VertexCount= 0;
	MJ_main.Junction.clear();

	MJ_main.EdgeCount= 0;
	MJ_main.EdgeHead.clear();
	MJ_main.EdgeTail.clear();

	//Select input file
	//-----------------
	std::string FileName;
	std::cout << "\nPlease enter your desired filename \n (extension .Sketch will be added automatically)\n" << std::endl;
	std::cin >> FileName;

	FileName += ".stk";
	std::ifstream in (FileName.c_str());

	//Read Vertices and Edges from input file
	//---------------------------------------
	if (in.is_open()){

		std::string line;
		double cont= -1;
		double stop= -1;
		while (getline(in, line)){
			if(line.substr(0, 16)== "number_of_points"){
				long Num= 17;
				while(line.substr(Num, 1)== " ")
					Num++;
				long EndNum= line.find(' ', Num);
				if(EndNum<0)
					EndNum= line.size();
				stop= atof(line.substr(Num, EndNum-Num).c_str());
				cont= -1;
			}
			if(((cont== 0)&&(stop!=-1))||(cont== stop-1)){
				POINT2D p;
				long BeginX= 0;
				long MidX= line.find('.', BeginX);
				long EndX= line.find(' ', MidX);
				p.x= atof(line.substr(BeginX, EndX-BeginX).c_str());

				long BeginY= EndX+1;
				while(line.substr(BeginY, 1) == " ")
					BeginY++;
				long EndY= line.find(' ', BeginY);
				if(EndY<0)
					EndY= line.size()-1;
				p.y= atof(line.substr(BeginY, EndY-BeginY).c_str());
				
				MJ_main.Junction.push_back(p);
					MJ_main.VertexCount++;
			
				if((MJ_main.VertexCount>0)&&(MJ_main.VertexCount%2== 0)){
					MJ_main.EdgeHead.push_back(MJ_main.VertexCount-2);
					MJ_main.EdgeTail.push_back(MJ_main.VertexCount-1);
					
					MJ_main.EdgeCount++;
					cont= -1;
					stop= -1;
				}
			}
			
			cont++;
		} //end while

		// Closes the file
		in.close();
	}//end if
	else{
		printf("Unable to open file\n");
	}
}


void Junctions2D::ResetMerits()
//Resets merits of vertices in the global database
{
	POINT2D vi;
	for (long i=0; i<(long) MJ_main.VertexCount; i++){
		vi= MJ_main.Junction[i];
		vi.merit= 1;
		MJ_main.SetPOINT2D(i, vi);
	}
}


void Junctions2D::ReorderLines()
//Reorders lines from longer to shorter
//and list them in "PriorLine"

//Also saves their lengths in "LenghtEdges"
{
	//Calculate lengths of lines
	for (long i=0; i<(long) MJ_main.EdgeCount; i++){
		LengthEdge.push_back(MJ_main.GetLength2D(i));
		PriorLine.push_back(i);
	}

	//Bubble sort lines
	std::vector<double> buble= LengthEdge;
	bool swapped;
	do{
		swapped= false;
		for (long i= 0; i<(long)MJ_main.EdgeCount-1; i++){
			if (buble[i] < buble[i+1]){  //from longer to shorter
				std::swap(buble[i], buble[i+1]);
				std::swap(PriorLine[i], PriorLine[i+1]);
				swapped= true;
			}
		}
	}while(swapped);
}


void Junctions2D::GetAllSemiLines()
//List all the semi-lines
{
	long numL= 0;

	for (long i=0; i<(long)MJ_main.EdgeCount; i++){  
		long pos= PriorLine[i]; //Longer lines are parsed first
		semiLines.push_back(std::vector<long>());
		semiLines[numL].push_back(PriorLine[i]);   //Number of the line
		semiLines[numL].push_back(MJ_main.EdgeHead[pos]);     //First tip
		semiLines[numL].push_back(MJ_main.EdgeTail[pos]);     //Second tip
		numL++;

		semiLines.push_back(std::vector<long>());
		semiLines[numL].push_back(PriorLine[i]);
		semiLines[numL].push_back(MJ_main.EdgeTail[pos]);
		semiLines[numL].push_back(MJ_main.EdgeHead[pos]);
		numL++;
	}
}


double Junctions2D::AvgLength(long numTip1, long numTip2)
//Returns the average length of the lines that contain tips numTip1 or numTip2
{
	double L= 0;
	long n= 0;
	for (long i=0; i<(long) MJ_main.EdgeCount; i++){
		long Head= MJ_main.EdgeHead[i];
		long Tail= MJ_main.EdgeTail[i];
		if(Head == numTip1 || Tail == numTip1 ||
		   Head == numTip2 || Tail == numTip2){
			L+= LengthEdge[i];
			n++;
		}
	}
	if(n==0)
		return 0;
	else
		return L/n;
}


long Junctions2D::GetFarTip(long CloseTip, long &line)
//Returns the tip connected to CloseTip

//Also returns the line that connects both tips

//WARNING: it is only valid for tips of valence 1!!
{
	for (long i=0; i<(long)semiLines.size(); i++){
		if(semiLines[i][1] == CloseTip){
			line= semiLines[i][0];
			return semiLines[i][2];
		}
	}
	return -1;
}


double Junctions2D::GetIntersection(double x1, double y1, double x11, double y11,
									double x2, double y2, double x22, double y22,
									double maxDist, double &x, double &y)
//Returns a merit value which is 1 if both tips are coincidental, otherwise
//it is in the range [0, 1] depending on the distance from the intersection point to the tips

//Lines may indisctintly intersect inside their segments,
//or outside the segments and by the side of their first tips
//In both cases the intersection must be closer than maxDist to the tips
//Returns 0 otherwise

//First line goes through points (x1,y1) and (x11,y11)
//Second line goes through points (x2,y2) and (x22,y22)

//Also returns the intersection point (x, y)

//Based on: Paul Bourke, April 1989 (http://paulbourke.net/geometry/lineline2d/)
{
	double denom  = (y22-y2) * (x11-x1) - (x22-x2) * (y11-y1);
	double numera = (x22-x2) * (y1-y2) - (y22-y2) * (x1-x2);
	double numerb = (x11-x1) * (y1-y2) - (y11-y1) * (x1-x2);

	if (fabs(numera) < Accuracy && fabs(numerb) < Accuracy && fabs(denom) < Accuracy)
		return -1;  //Lines are coincident

	if (fabs(denom) < Accuracy)
		return -1;  //lines are parallel

	//Find the intersection
	double mua = numera / denom;
	double mub = numerb / denom;

	x = x1 + mua * (x11 - x1);
	y = y1 + mua * (y11 - y1);

	return (1. - ((abs(mua)+abs(mub))/2.)/maxDist);   //The merit depends on the average edge
}

double Junctions2D::MeritRotation(long NumTip, POINT2D newPoint)
//Returns the merit of tip NumTip
//which decreases when the lines connected to it
//need to be rotated when redirected to newPoint

//Merit is 1 if lines do not rotate,
//and is negative if at least one of them rotates more than maxRotation
{
	double meritR= 1.;

	if(maxRotation >= NUMBER_PI/2)
		return meritR;

	POINT2D v= MJ_main.Junction[NumTip];
	POINT2D t;
	t.assign(v.x, v.y);

	std::vector<long> edgesInVertex= MJ_main.GetAllEdgesInVertex(NumTip);
	for (long i=0; i<(long)edgesInVertex.size(); i++){
		long pos= edgesInVertex[i];
		POINT2D p;
		if (MJ_main.EdgeHead[pos] == NumTip)
			p= MJ_main.Junction[MJ_main.EdgeTail[pos]];
		else if (MJ_main.EdgeTail[pos] == NumTip)
			p= MJ_main.Junction[MJ_main.EdgeHead[pos]]; 
		POINT2D pivot;
		pivot.assign(p.x, p.y);

		meritR= min(meritR, (1. - (GetAngle_PI(t, pivot, newPoint) / maxRotation)));
	}

	return meritR;
}


double Junctions2D::MeritTips(long NumTip1, long NumTip2,
							  POINT2D &junction, bool &swapTips)
//Calculate merit of merging tips NumTip1 and NumTip2

//The junction is also calculated, as follows:
//   - If both tips belong to edges not connected (their valences are 1),
//     then the merged tip is placed in their intersection

//   - If both tips have the same valence (and this valence is different from 1),
//     the midpoint of the segment connecting their closer tips 
//     is used as junction

//   - If, at least, one tip is shared by more than one edge,
//     then the vertex with the highest valence (the one where more edges end)
//     becomes the junction


//The approach uses an allowance (based on a cosine function) to prioritize perpendicular lines:
//   - Collinear lines are given a 100% allowance
//     while perpendicular lines are given a 200% allowance
//   - Transition is trigonometric, instead of linear, to give a great allowance for
//     lines in the range of "axonometric angles" (between 60 and 90 degrees)
{
	//Get coordinates and valences of the tips
	POINT2D v1= MJ_main.Junction[NumTip1];
	POINT2D t1;
	t1.assign(v1.x, v1.y);

	POINT2D v2= MJ_main.Junction[NumTip2];
	POINT2D t2;
	t2.assign(v2.x, v2.y);

	long Valence1= MJ_main.GetSizeEdgesInVertex(NumTip1);
	long Valence2= MJ_main.GetSizeEdgesInVertex(NumTip2);

	swapTips= false;

	double merit;

	//Merit of isolated lines
	//-----------------------
	if (Valence1==1 && Valence2==1){
		long line1, line2;
		POINT2D t11= MJ_main.Junction[GetFarTip(NumTip1, line1)];  //far tip of the first semi-line
		POINT2D t22= MJ_main.Junction[GetFarTip(NumTip2, line2)];  //far tip of the second semi-line

		//Calculate allowance, to prioritize perpendicular lines
		POINT2D segment1;
		segment1.assign(t11.x-t1.x, t11.y-t1.y);
		POINT2D segment2;
		segment2.assign(t22.x-t2.x, t22.y-t2.y); 

		double allowance= (2. - abs(DOT_PRODUCT_2D(segment1, segment2))/LengthEdge[line1]/LengthEdge[line2]);
		double maxDistAlw= maxDistance * allowance;

		//Get merit and intersection point
		merit= GetIntersection(t1.x, t1.y, t11.x, t11.y,
							   t2.x, t2.y, t22.x, t22.y,
							   maxDistAlw,
							   junction.x, junction.y);
	}

	//Merit of already connected lines
	//--------------------------------
	else{
		double meritD, meritR;
		meritD= (1. - (MODULE_2D(t1, t2) / (AvgLength(NumTip1, NumTip2) * maxDistance)));

		if(Valence2 == Valence1){
			//Place junction at the midpoint
			junction.x= (t1.x+t2.x)/2;
			junction.y= (t1.y+t2.y)/2;
			meritR= min(MeritRotation(NumTip1, junction), MeritRotation(NumTip2, t1));
		}

		else if(Valence2 > Valence1){
			//Swap tips to place the junction at the tip with the highest valence
			swapTips= true;
			junction= t2;
			meritR= MeritRotation(NumTip2, t1);
		}

		else{
			//Place the junction at the tip with the highest valence
			junction= t1;
			meritR= MeritRotation(NumTip2, t1);
		}

		merit= (meritD +    //Extra merit if tips are closer to each other than maxDistance
			    meritR)/2;  //Extra merit if the edges connected to the vertex rotate less than maxRotation
	}

	//Return merit
	//------------
	if(merit >= 0)
		return (1-ReificationMerit) * merit + ReificationMerit;  //Add reward for careful vertices
	else
		return -1.;   //Do not merge if tips are too distant, and/or lines rotate too much
}


void Junctions2D::DeleteNeighbor(long numNeighbor)
//Deletes the neighbor tip numNeighbor,
//and renumbers all subsequent neighbors
{
	Neighbor.erase(Neighbor.begin()+numNeighbor);

	for (long i= 0; i<(long)Neighbor.size(); i++){
		for (long j= 0; j<(long)Neighbor[i].size(); j++){
			if(Neighbor[i][j] == numNeighbor){
				Neighbor[i].erase(Neighbor[i].begin()+j);
				j--;
			}
			else if(Neighbor[i][j] > numNeighbor){
				Neighbor[i][j]--;
			}
		}
	}
}


void Junctions2D::MergeTips(long NumTip1, long NumTip2,
							POINT2D junction, double meritT)
//Updates the global database by reconecting to vertex NumTip1
//all the edges connected to vertex NumTip2,
//then removes vertex NumTip2

//Also updates the semi-lines list, and the Neighbor list

//We note that the far vertices of the merged lines
//are removed from the neighbor list
//as the can no longer be neighbor of the merged vertex
{
	//Update vertex 1
	POINT2D vertex1= MJ_main.Junction[NumTip1];
	vertex1.x= junction.x;
	vertex1.y= junction.y;
	vertex1.merit= min(vertex1.merit, meritT);
	MJ_main.SetPOINT2D(NumTip1, vertex1);

	//Re-connect to vertex 1 all the edges connected to vertex 2
	std::vector<long> edgesInVertex= MJ_main.GetAllEdgesInVertex(NumTip2);
	for (long k=0; k<(long)edgesInVertex.size(); k++){
		long pos= edgesInVertex[k];

		long NumTip22;
		if (MJ_main.EdgeHead[pos] == NumTip2){
			MJ_main.EdgeHead[pos] =NumTip1;
			NumTip22= MJ_main.EdgeTail[pos];
		}
		else if (MJ_main.EdgeTail[pos] == NumTip2){
			MJ_main.EdgeTail[pos] = NumTip1;
			NumTip22= MJ_main.EdgeHead[pos];
		}

		//Remove the far vertex from the neighbor list
		for (long j= 0; j<(long)Neighbor[NumTip1].size(); j++){
			if(Neighbor[NumTip1][j] == NumTip22){
				Neighbor[NumTip1].erase(Neighbor[NumTip1].begin()+j);
				break;
			}
		}
		for (long j= 0; j<(long)Neighbor[NumTip22].size(); j++){
			if(Neighbor[NumTip22][j] == NumTip1){
				Neighbor[NumTip22].erase(Neighbor[NumTip22].begin()+j);
				break;
			}
		}
	}

	//Delete vertex 2 (and update the edges)
	MJ_main.RemovePOINT2D(NumTip2, true);

	//Update the list of semi-lines
	for (long i=0; i<(long)semiLines.size(); i++){
		if(semiLines[i][1] == NumTip2) semiLines[i][1]= NumTip1;
		if(semiLines[i][1] > NumTip2) semiLines[i][1]--;
		if(semiLines[i][2] == NumTip2) semiLines[i][2]= NumTip1;
		if(semiLines[i][2] > NumTip2) semiLines[i][2]--;
	}

	//Delete the removed vertex from the neighbor list
	DeleteNeighbor(NumTip2);
}

long Junctions2D::MergeCarefulTips()
//Merge neighbor tips of lines into common junctions that represent vertices

//This is a Nearest-Neighbor approach, as it is based on the assumption that
//tips of carefully drawn lines that share junctions must be neighbor
//from each other in a Delaunay triangulation of the cloud of tips

//The function merges sequentially the pairs of tips that best fit
//the following merging conditions:
//		- They are neighbor and belong to different lines
//		- Are closer than a threshold distance "maxDistance",
//		  which varies depending on the angle of their lines(*)
//		- Lines to connect rotate less than a threshold rotation "maxRotation"
//(*) maxDistance is weighted to prioritize perpendicular lines, as 
//    an allowance is used to prioritize lines that meet at a square angle

//Lines are parsed from longer to shorter,
//as it is asumed that short lines define auxiliary elements

//The Delaunay triangulation was adapted from:
//		Gilles Dumoulin's C++ implementation of Paul Bourke's Delaunay triangulation algorithm.
//		Source: http://paulbourke.net/papers/triangulate/


//Implemented by Pedro Company, 2017
{
	GetAllSemiLines();

	long DeletedVertices= 0;
	std::vector<bool> Merged(MJ_main.VertexCount,false);

	for(long i=0; i<(long)semiLines.size(); i++){
		long tipI= semiLines[i][1];

		if(Merged[tipI])
			continue;

		bool merging;
		do{
			merging= false;

			long bestJ;
			double bestMerit= -1.;
			bool bestSwap= false;
			POINT2D bestJunction;

			for(long j= 0; j<(long)Neighbor[tipI].size(); j++){
				long tipJ= Neighbor[tipI][j];  //J-th neighbor tip of tipI

				if(Merged[tipJ])
					continue;  //The closer tip is already merged
				if(tipJ == tipI)
					continue;  //Both tips were previously merged

				//Calculate the merit of the candidate junction
				POINT2D junction;
				bool swapTips;
				double merit= MeritTips(tipI, tipJ, junction, swapTips);

				//Save current best junction
				if(merit > bestMerit){
					bestJ= tipJ;
					bestMerit= merit;
					bestSwap= swapTips;
					bestJunction= junction;
				}
			}

			//Merge the best valid junction
			if(bestMerit >= 0){
				if(bestSwap){
					MergeTips(bestJ, tipI, bestJunction, bestMerit);  //Also updates semi-lines and neighbors
					Merged[bestJ]= true;
					Merged.erase(Merged.begin()+tipI);
					tipI= bestJ;  //tipI has been renumbered
				}
				else{
					MergeTips(tipI, bestJ, bestJunction, bestMerit);  //Also updates semi-lines and neighbors
					Merged[tipI]= true;
					Merged.erase(Merged.begin()+bestJ);
					if(bestJ < tipI)
						tipI--; //tipI has been renumbered
				}
				DeletedVertices++;
				merging= true;
			}
		}while(merging);
	}

	semiLines.clear();

	return DeletedVertices;
}


void Junctions2D::GetTripletSemiLines()
//List all the dangling semi-lines
//(These are lines whose tip is not connected to any other line)

//They are listed to search for triplets of lines that intersect each other
{
	//Calculate valences of vertices, to detect dangling semi-lines
	std::vector<long> val;
	for (long i=0; i<(long)MJ_main.VertexCount; i++){
		val.push_back(MJ_main.GetSizeEdgesInVertex(i));
	}

	//List all the dangling semi-lines
	for (long i=0; i<(long)MJ_main.EdgeCount; i++){ 
		long pos= PriorLine[i];//Longer lines are parsed first
		if(val[MJ_main.EdgeHead[pos]]==1){
			SLX slx;
			slx.sl.push_back(PriorLine[i]);
			slx.sl.push_back(MJ_main.EdgeHead[pos]);
			slx.sl.push_back(MJ_main.EdgeTail[pos]);
			sLines.push_back(slx);
		}
		if(val[MJ_main.EdgeTail[pos]]==1){
			SLX slx;
			slx.sl.push_back(PriorLine[i]);
			slx.sl.push_back(MJ_main.EdgeTail[pos]);
			slx.sl.push_back(MJ_main.EdgeHead[pos]);
			sLines.push_back(slx);
		}
	}
}


void Junctions2D::RemoveDanglingSemiLine(long num)
//Removes one dangling semi-line
//and updates the numbers of the remaining semi-lines
{
	sLines.erase(sLines.begin()+num);

	//Update numbers of semi-lines
	//and remove references to the dangling line removed
	for (long j=0; j<(long)sLines.size(); j++){
		for (long k=0; k<(long)sLines[j].lineX.size(); k++){
			if(sLines[j].lineX[k] > num)
				//Update number of a subsequent semi-line
				sLines[j].lineX[k]--;
			else if(sLines[j].lineX[k] == num){
				//Remove reference to the dangling line removed
				sLines[j].lineX.erase(sLines[j].lineX.begin()+k);
				sLines[j].pointX.erase(sLines[j].pointX.begin()+k);
				k--;
			}
		}
	}
}


bool Junctions2D::SemiLinesIntersect(double x1, double y1, double x11, double y11,
									 double x2, double y2, double x22, double y22,
									 double &ta, double &tb, POINT2D &px)
//Returns true if semi-lines intersect in the valid range,
//which covers an asimetric region close to the first tip (both inside and outside the segments)
//Returns false otherwise

//First line contains the segment defined by tips (x1,y1) and (x11,y11)
//Second line contains the segment defined by tips (x2,y2) and (x22,y22)

//Also returns:
//    the parametric distance from the tips to the intersection point (ta, tb)
//    the intersection point (x, y)

//The parametric distances "t" from tips to the intersection are:
//   0 if intersection is in the tip,
//   1 if intersection is in the other tip of the same line, and
//  -1 if intersection is outside the segment and at a distance equal to the lenght of the tip
//The valid range for distances "t" is [CasualInTol, CasualOutTol]

//Based on: Paul Bourke, April 1989 (http://paulbourke.net/geometry/lineline2d/)
{
	double denom  = (y22-y2) * (x11-x1) - (x22-x2) * (y11-y1);
	double numera = (x22-x2) * (y1-y2) - (y22-y2) * (x1-x2);
	double numerb = (x11-x1) * (y1-y2) - (y11-y1) * (x1-x2);

	if (fabs(numera) < Accuracy && fabs(numerb) < Accuracy && fabs(denom) < Accuracy)
		return false;  //Lines are coincident

	if (fabs(denom) < Accuracy)
		return false;  //lines are parallel

	//Find the intersection
	ta = numera / denom;
	tb = numerb / denom;

	//Check if intersection is in the valid ranges
	if (ta > CasualInTol || ta < -CasualOutTol ||   //Intersection is in the valid range of the first line
		tb > CasualInTol || tb < -CasualOutTol)     //Intersection is in the valid range of the second line
		return false; //The intersection is far from, at least, one tip

	px.x = x1 + ta * (x11 - x1);
	px.y = y1 + ta * (y11 - y1);

	return true;
}

void Junctions2D::GetTripletsIntersections()
//The dangling semi-lines that intersect each semi-line I (if any) are tabulated in lineX
//Semi-lines in lineX are ordered from closer to farther to the first tip of line I

//Intersection points between line I and lines X are also saved (pointX)
{
	//Reset parametric distances used to order intersections
	std::vector<std::vector<double>> distTo;
	for (long i=0; i<(long)sLines.size(); i++){
		distTo.push_back(std::vector<double>());
	}

	long numP= 0;

	for (long i=0; i<(long)sLines.size(); i++){
		POINT2D vi, vii;
		vi= MJ_main.Junction[sLines[i].sl[1]];
		vii= MJ_main.Junction[sLines[i].sl[2]];

		//Loop for the intersecting dangling lines
		for (long j=i+1; j<(long)sLines.size(); j++){
			POINT2D vj, vjj;
			vj= MJ_main.Junction[sLines[j].sl[1]];
			vjj= MJ_main.Junction[sLines[j].sl[2]];

			//Save dangling semi-lines that intersect semi-line I inside the valid range
			double tti, ttj;
			POINT2D px;
			if(SemiLinesIntersect(vi.x, vi.y, vii.x, vii.y,
								  vj.x, vj.y, vjj.x, vjj.y,
								  tti, ttj, px)){

				if(tti < 0)
					distTo[i].push_back(-CasualInTol/CasualOutTol*tti);  //Increase distance for lines that intersect outside
				else
					distTo[i].push_back(tti);

				sLines[i].lineX.push_back(j);
				sLines[i].pointX.push_back(px);

				if(ttj < 0)
					distTo[j].push_back(-CasualInTol/CasualOutTol*ttj);  //Increase distance for lines that intersect outside
				else
					distTo[j].push_back(ttj);

				sLines[j].lineX.push_back(i);
				sLines[j].pointX.push_back(px);
			}
		}
	}

	//Bubble sort intersecting lines from closest to farther from the tip of each semi-line
	for (long i=0; i<(long)sLines.size(); i++){
		bool swapped;
		do{
			swapped= false;
			for (long k= 0; k<(long)distTo[i].size()-1; k++){
				if (distTo[i][k] > distTo[i][k+1]){
					std::swap(distTo[i][k], distTo[i][k+1]);
					std::swap(sLines[i].lineX[k], sLines[i].lineX[k+1]);
					std::swap(sLines[i].pointX[k], sLines[i].pointX[k+1]);
					swapped= true;
				}
			}
		}while(swapped);
	}

	//Remove dangling lines not in range
	for (long i=0; i<(long)sLines.size(); i++){
		if(sLines[i].lineX.size() < 2){       //of at least other two intersecting lines
			RemoveDanglingSemiLine(i);
			i--;
		}
	}
}


void Junctions2D::RenumberTripletTips(long v1, long v2, long v3)
//Updates the local database by renumbering the merged tips

//Renumbers tips v2 and v3 as v1
//and reduces the numbers subsequent to v2 and v3
{
	if(v2 > v3)
		std::swap(v2, v3);

	for (long i=0; i<(long)sLines.size(); i++){
		if(sLines[i].sl[1] == v2 || sLines[i].sl[1] == v3)
			sLines[i].sl[1]= v1;
		else{
			if(sLines[i].sl[1] > v2 && sLines[i].sl[1] <= v3)
				sLines[i].sl[1]--;
			else if(sLines[i].sl[1] > v3)
				sLines[i].sl[1]-=2;
		}

		if(sLines[i].sl[2] == v2 || sLines[i].sl[2] == v3)
			sLines[i].sl[2]= v1;
		else{
			if(sLines[i].sl[2] > v2 && sLines[i].sl[2] <= v3)
				sLines[i].sl[2]--;
			else if(sLines[i].sl[2] > v3)
				sLines[i].sl[2]-=2;
		}
	}
}



void Junctions2D::MergeTriplet(long numDangling, long numNeighbor, POINT2D center, double merit,
							   long line2, long line3)
//Defines a common junction in the global database,
//shared by lines line1, line2 and line3
{
	//Define vertices to merge
	long v1= sLines[numDangling].sl[1];
	long v2= sLines[sLines[numDangling].lineX[numNeighbor]].sl[1];
	long v3= sLines[sLines[numDangling].lineX[numNeighbor+1]].sl[1];

	//Update vertex 1 as common vertex
	center.merit= EmergentTripletsMerit * merit;
	MJ_main.SetPOINT2D(v1, center);

	//Reconect vertices 2 and 3 to vertex1
	if(v2==MJ_main.EdgeHead[line2])
		MJ_main.EdgeHead[line2]= v1;
	else
		MJ_main.EdgeTail[line2]= v1;

	if(v3==MJ_main.EdgeHead[line3])
		MJ_main.EdgeHead[line3]= v1;
	else
		MJ_main.EdgeTail[line3]= v1;

	//Remove old vertices and renumber
	if(v3 > v2){
		MJ_main.RemovePOINT2D(v3, true);
		MJ_main.RemovePOINT2D(v2, true);
	}
	else{
		MJ_main.RemovePOINT2D(v2, true);
		MJ_main.RemovePOINT2D(v3, true);
	}

	//Renumber tips in the local database of dangling semi-lines
	RenumberTripletTips(v1, v2, v3);

	//Remove the triplet from the lists of intersections of dangling semi-lines
	long dangling1= numDangling;
	long dangling2= sLines[numDangling].lineX[numNeighbor];
	long dangling3= sLines[numDangling].lineX[numNeighbor+1];

	RemoveDanglingSemiLine(dangling1);
	if(dangling2 > dangling1) dangling2--;
	if(dangling3 > dangling1) dangling3--;

	RemoveDanglingSemiLine(dangling2);
	if(dangling3 > dangling2) dangling3--;

	RemoveDanglingSemiLine(dangling3);
}


long Junctions2D::GetTriplets()
//Triplets of dangling semi-lines are connected to define a commmon junction
//if their three intersection points are close to each other, and close to their tips

//Only parses the two first semi-lines that intersect each semi-line
//(those that are closer to the tips)
{
	long NumRemovedTips= 0;

	//Loop to search for triplets of semi-lines that intersect close to each other
	bool triplets;
	do{
		triplets= false;

		//Check the longest unvisited dangling semi-line
		for (long i=0; i<(long)sLines.size(); i++){
			//Check if the two first lines that intersect the current semi-line
			//are the first to intersect to each other
			if((sLines[i].lineX.size() > 1 &&
				sLines[sLines[i].lineX[0]].lineX.size() > 1 &&
				sLines[sLines[i].lineX[1]].lineX.size() > 1)
				    &&
				    ((sLines[sLines[i].lineX[0]].lineX[0] == i                  && sLines[sLines[i].lineX[0]].lineX[1] == sLines[i].lineX[1]) ||
				     (sLines[sLines[i].lineX[0]].lineX[0] == sLines[i].lineX[1] && sLines[sLines[i].lineX[0]].lineX[1] == i))
				    &&
				    ((sLines[sLines[i].lineX[1]].lineX[0] == i                  && sLines[sLines[i].lineX[1]].lineX[1] == sLines[i].lineX[0]) ||
				     (sLines[sLines[i].lineX[1]].lineX[0] == sLines[i].lineX[0] && sLines[sLines[i].lineX[1]].lineX[1] == i))
					){

				//Calculate tolerance for the triplet
				long line2= sLines[sLines[i].lineX[0]].sl[0];
				long line3= sLines[sLines[i].lineX[1]].sl[0];

				double maxDistTripletI= (LengthEdge[sLines[i].sl[0]] +
										 LengthEdge[line2] +
										 LengthEdge[line3]) /3 *
										 maxDistanceTriplets;

				//Get intersections
				POINT2D X1= sLines[i].pointX[0];
				POINT2D X2= sLines[i].pointX[1];
				POINT2D X3= sLines[sLines[i].lineX[0]].pointX[0];
				if(X3==X1 || X3==X2)
					X3= sLines[sLines[i].lineX[0]].pointX[1];

				//Check if intersections are in range
				if(MODULE2D(X1.x, X1.y, X2.x, X2.y) < maxDistTripletI &&
				   MODULE2D(X2.x, X2.y, X3.x, X3.y) < maxDistTripletI &&
				   MODULE2D(X3.x, X3.y, X1.x, X1.y) < maxDistTripletI){

					//Calculate merit
					POINT2D center;
					center.x=(X1.x + X2.x + X3.x)/3;
					center.y=(X1.y + X2.y + X3.y)/3;

					double distCentroid= max(MODULE2D(center.x, center.y, X1.x, X1.y),
										 max(MODULE2D(center.x, center.y, X2.x, X2.y),
										 MODULE2D(center.x, center.y, X3.x, X3.y)));

					double merit= 1. - distCentroid/maxDistTripletI;

					//Merge the tips if the merit is non-negative
					if(merit >= 0){
						//Define a common junction in the database
						MergeTriplet(i, 0, center, merit, line2, line3);

						NumRemovedTips+= 2;
						triplets= true;
					}
					break;
				}
			}
		}
	}while(triplets);

	return NumRemovedTips;
}

long Junctions2D::GetPoorTriplets()
//Triplets of dangling lines are connected to define a commmon junction
//if their three intersection points are close to each other
{
	long NumRemovedTips= 0;

	//Check the longest unvisited dangling semi-line
	for (long i=0; i<(long)sLines.size(); i++){

		//Search for close intersecting lines
		long bestj;
		double bestMerit= -1.;
		POINT2D bestCenter;
		long bestLine2, bestLine3;

		long jsize= sLines[i].lineX.size();
		if(jsize > 1) for (long j=0; j<jsize-1; j++){

			long ksize= sLines[sLines[i].lineX[j]].lineX.size();
			if(ksize > 1) for (long k=0; k<ksize-1; k++){

				long msize= sLines[sLines[i].lineX[j+1]].lineX.size();
				if(msize > 1) for (long m=0; m<msize-1; m++){

					if(((sLines[sLines[i].lineX[j]].lineX[k] == i                     && sLines[sLines[i].lineX[j]].lineX[k+1] == sLines[i].lineX[j+1]) ||
						(sLines[sLines[i].lineX[j]].lineX[k] == sLines[i].lineX[j+1] && sLines[sLines[i].lineX[j]].lineX[k+1] == i))
					   &&
					   ((sLines[sLines[i].lineX[j+1]].lineX[m] == i                  && sLines[sLines[i].lineX[j+1]].lineX[m+1] == sLines[i].lineX[j]) ||
						(sLines[sLines[i].lineX[j+1]].lineX[m] == sLines[i].lineX[j] && sLines[sLines[i].lineX[j+1]].lineX[m+1] == i))
					  ){

						//Calculate tolerance for the triplet
						long line1= sLines[i].sl[0];
						long line2= sLines[sLines[i].lineX[j]].sl[0];
						long line3= sLines[sLines[i].lineX[j+1]].sl[0];

						double maxDistTripletI= (LengthEdge[line1] +
												 LengthEdge[line2] +
												 LengthEdge[line3]) /3 *
												 maxDistanceTriplets;

						//Get intersections
						POINT2D X1= sLines[i].pointX[j];
						POINT2D X2= sLines[i].pointX[j+1];
						POINT2D X3= sLines[sLines[i].lineX[j]].pointX[0];
						if(X3==X1 || X3==X2)
							X3= sLines[sLines[i].lineX[j]].pointX[1];

						//Check if intersections are in range
						if(MODULE2D(X1.x, X1.y, X2.x, X2.y) < maxDistTripletI &&
						   MODULE2D(X2.x, X2.y, X3.x, X3.y) < maxDistTripletI &&
						   MODULE2D(X3.x, X3.y, X1.x, X1.y) < maxDistTripletI){

							//Calculate merit
							POINT2D center;
							center.x=(X1.x + X2.x + X3.x)/3;
							center.y=(X1.y + X2.y + X3.y)/3;

							double distCentroid= max(MODULE2D(center.x, center.y, X1.x, X1.y),
												 max(MODULE2D(center.x, center.y, X2.x, X2.y),
												 MODULE2D(center.x, center.y, X3.x, X3.y)));

							double merit= 1. - distCentroid/maxDistTripletI;

							//Save the triplet with the best merit
							if(merit > bestMerit){
								bestj= j;
								bestMerit= merit;
								bestCenter= center;
								bestLine2= line2;
								bestLine3= line3;
							}
						}
					}
				}
			}
		}

		//Merge the best valid triplet
		if(bestMerit >= 0){
			long j= sLines[i].lineX[bestj];
			long j1= sLines[i].lineX[bestj+1];

			MergeTriplet(i, bestj, bestCenter, bestMerit, bestLine2, bestLine3);

			i--;
			if(i > j) i--;
			if(i > j1) i--;

			NumRemovedTips+= 2;
		}
	}

	return NumRemovedTips;
}


long Junctions2D::mergeCasualTriplets()
//Triplets of dangling lines are connected to define a commmon junction
//if their three intersection points are close to each other

//Lines are parsed as two "semi-lines":
//first and second tips are switched to get each semi-line
//A semi-line is dangling if its first tip is not shared
//by any other line (it is said that the "valence" of this tip is 1)

//The procedure is as follows:
//   1. All the dangling semi-lines are listed
//   2. The dangling semi-lines that intersect each semi-line I in the valid range (if any) are listed (lineX)
//...3. Semi-lines X are ordered from closest to farthest to the tip of semi-line I
//   4. If the two semilines that are close to the tip of a given semiline
//      are also close to their own tips, the triplet is selected
//   5  Then it is checked whether its mutual intersections are also close to each other
//...5. The three semi-lines that intersect close to each other are connected to a common junction
//      located in the centroid of the three intersection points
//   6. The merit of the common junction is assigned as
//      EmergentTripletsMerit*(1-distCentroid/maxDistTriplets),
//      where distCentroid is the distance between the centroid and the most distant intersection point
//   7. The procedure is repeated while valid triplets are found
//Longer semi-lines are parsed first
{
	long NumRemovedTips= 0;

	//List all the dangling semi-lines
	GetTripletSemiLines();

	//List all the valid intersections between dangling semi-lines
	//(and removes isolated dangling lines)
	GetTripletsIntersections();

	//Loop to search for triplets of lines that intersect close to each other
	//and close to the tips
	NumRemovedTips+= GetTriplets();

	//Loop to search for other (poor quality) triplets
	NumRemovedTips+= GetPoorTriplets();

	//Clear list of dangling semilines
	sLines.clear();

	//Clear list of lengths
	LengthEdge.clear();

	return NumRemovedTips;
}

void Junctions2D::GetDanglingSemiLines()
{
	//Calculate valences, to detect dangling lines
	for (long i=0; i<(long)MJ_main.VertexCount; i++){
		valence.push_back(MJ_main.GetSizeEdgesInVertex(i));
	}

	//Get dangling semilines
	long numL= 0;

	for (long i=0; i<(long)MJ_main.EdgeCount; i++){
		long pos= PriorLine[i];//Longer lines are parsed first

		if(valence[MJ_main.EdgeHead[pos]]==1){
			dsLines.push_back(std::vector<long>());
			dsLines[numL].push_back(PriorLine[i]);
			dsLines[numL].push_back(MJ_main.EdgeHead[pos]);
			dsLines[numL].push_back(MJ_main.EdgeTail[pos]);
			numL++;
		}

		if(valence[MJ_main.EdgeTail[pos]]==1){
			dsLines.push_back(std::vector<long>());
			dsLines[numL].push_back(PriorLine[i]);
			dsLines[numL].push_back(MJ_main.EdgeTail[pos]);
			dsLines[numL].push_back(MJ_main.EdgeHead[pos]);
			numL++;
		}
	}
}


double Junctions2D::alwDangling(POINT2D p1, POINT2D p11, long junction)
//Calculate allowance, to prioritize perpendicular lines

//The allowance is calculated between the dangling line (p1-p11)
//and its most collinear line in the junction
{
	double allowance= 2.;   //max allowance, for normal lines

	POINT2D segment1;
	segment1.x= p11.x-p1.x;
	segment1.y= p11.y-p1.y;
	double length1= MODULE_2D(p1, p11);

	//Loop to calculate the smallest allowance
	std::vector<long> edgesInVertex= MJ_main.GetAllEdgesInVertex(junction);
	for (long j=0; j<(long)edgesInVertex.size(); j++){
		long pos= edgesInVertex[j];
		POINT2D pj, pjj;
		if (MJ_main.EdgeHead[pos] == junction){
			pj= MJ_main.Junction[MJ_main.EdgeHead[pos]];
			pjj= MJ_main.Junction[MJ_main.EdgeTail[pos]];
		}
		else{
			pj= MJ_main.Junction[MJ_main.EdgeTail[pos]];
			pjj= MJ_main.Junction[MJ_main.EdgeHead[pos]];
		}

		POINT2D segment2;
		segment2.x= pjj.x-pj.x;
		segment2.y= pjj.y-pj.y;
		double length2= MODULE_2D(pj, pjj);

		double allowancej= (2. - abs(DOT_PRODUCT_2D(segment1, segment2))/length1/length2);

		if(allowancej < allowance)
			allowance= allowancej;
	}

	return allowance;
}

double Junctions2D::meritCasualDangling(POINT2D p1, POINT2D p11, double length,
										long junction)
//Returns the merit of merging the dangling segment p1-p11
//and the junction, which must be smaller than maxDistanceDangling
//while its closest point must be in range

//The merit of the junction is linearly reduced

//One merit equals 1 if the dangling line passes through the junction,
//or 0 if the dangling line passes at a distance of maxDistanceDangling or greater

//Another merit is calculated depending on the distance between the tip of the semi-line
//and the point of perpendicular projection of the junction on the line

//The minimum of both merits is returned as merit of the close junction

//The calculation of the minimum distance is based on:
	//Given the point C (Cx,Cy) and the segment AB (Ax,Ay) to (Bx,By).
	//Let P be the point of perpendicular projection of C on AB
	//The parameter t, which indicates P's position along AB,
	//is computed by the dot product of AC and AB
	//divided by the square of the length of AB
	//Philip Nicoletti (2002) CodeGuru
{

	POINT2D junc= MJ_main.Junction[junction];

	double t = ((junc.x - p1.x) * (p11.x - p1.x) + (junc.y - p1.y) * (p11.y - p1.y)) / (length*length);

	//Check if the distance normal intersects in the valid range
	if (t > CasualInTol ||  //The intersection is valid if it occurs in the half line close to the first tip
		t < -CasualOutTol)  //The intersection is also valid if it occurs in the near extension of the segment
		return -1;

	//Distance
	double dX = junc.x - (p1.x + t * (p11.x - p1.x));
	double dY = junc.y - (p1.y + t * (p11.y - p1.y));
	double dist= sqrt(dX*dX + dY*dY);

	//Check if the distance is in the valid range
	if (dist/length > CasualInTol)
		return -1;

	//Merits
	double merit1= (1 - (dist/(length*maxDistanceDangling*alwDangling(p1, p11, junction))));

	double merit2;
	if(t >= 0)
		merit2= (1 - t/CasualInTol);
	else
		merit2= (1 - t/(-CasualOutTol));

	return min(merit1, merit2);
}

long Junctions2D::mergeCasualDanglingLines()
//Tips of dangling semi-lines are connected to the closest junction
//if they are closer than a threshold distance

//This is based on the assumption that the line-drawing depicts a polyhedral shape,
//thus dangling lines are considered uncommon, and are merged if they pass near enough to a junction

//The procedure is as follows:
//    1.One dangling semi-line is selected (I)
//    2.For every junction J (tip of valence higher than 1)
//      closer than a threshold (maxDistDangling)
//      and within the valid range of semi-line I,
//      the merit of merging I to J is calculated
//    3.The junction JJ with the higher positive merit is selected
//    4.The tip of the dangling line I is added to the junction JJ,
//      only if the reduced merit of JJ is non-negative
//    5.The procedure is repeated for all dangling semi-lines
{
	long NumRemovedTips= 0;

	//Get dangling semilines
	GetDanglingSemiLines();

	//Loop to merge dangling semi-lines
	for(long i=0; i<(long)dsLines.size(); i++){
		POINT2D vi= MJ_main.Junction[dsLines[i][1]];
		POINT2D di;
		di.assign(vi.x, vi.y);
		POINT2D vii= MJ_main.Junction[dsLines[i][2]];
		POINT2D dii;
		dii.assign(vii.x, vii.y);

		double lengthi= MODULE_2D(di, dii);
		if (lengthi == 0)
			continue;

		//Loop to search for the junction most meritorious to merge
		double maxmerit= -1;
		long junCloser;

		for (long j=0; j<(long) MJ_main.VertexCount; j++){
			if(valence[j] < 2)
				continue;  //Do not merge danglindg lines to another dangling line
			if(j==dsLines[i][1] || j==dsLines[i][2])
				continue;  //Do not merge danglindg lines to themselves

			double merit= meritCasualDangling(di, dii, lengthi, j);

			if(merit > maxmerit){
				maxmerit= merit;
				junCloser= j;
			}
		}

		//Connect dangling semi-line to its closest junction
		if(maxmerit >= 0){
			//Update merit of the junction
			POINT2D Junction= MJ_main.Junction[junCloser];
			double newMerit= min(Junction.merit, maxmerit);

			if(newMerit >= 0){
				Junction.merit= newMerit;
				MJ_main.SetPOINT2D(junCloser, Junction);

				//Remove dangling tip
				long position= dsLines[i][0];
				if(MJ_main.EdgeHead[position] == dsLines[i][1])
					MJ_main.EdgeHead[position] = junCloser;
				else
					MJ_main.EdgeTail[position] = junCloser;

				MJ_main.RemovePOINT2D(dsLines[i][1], true);

				//Update pointers and flags
				valence[junCloser]++;
				valence.erase(valence.begin() + dsLines[i][1]);
				NumRemovedTips++;

				//Update the list of semi-lines
				long tipD= dsLines[i][1];
				for (long i=0; i<(long)dsLines.size(); i++){
					if(dsLines[i][1] == tipD) dsLines[i][1]= junCloser;
					if(dsLines[i][1] > tipD) dsLines[i][1]--;
					if(dsLines[i][2] == tipD) dsLines[i][2]= junCloser;
					if(dsLines[i][2] > tipD) dsLines[i][2]--;
				}
			}
		}
	}

	return NumRemovedTips;
}

long Junctions2D::GetJunctions(std::vector<POINT2D> &Vertices, std::vector<long> &HeadEdges, std::vector<long> &TailEdges)
//Merge tips of lines into common 2D junctions
//that are assumed to represent 3D vertices

//The goal is merging firstly those junctions most clearly perceived,
//while giving a second chance for poorly drawn junctions

//The approach follows three stages:
//		- Firstly, Neighbor tips are merged, iff they fit a threshold distance and an orientation limit
//		- Secondly, triplets of dangling lines are merged (if they intersect close to each other)
//		- Finally, isolated dangling lines are merged (if they pass close to a junction)

//The first stage is intended to reificate careful vertices,
//while the second and the third stages are aimed at emerging casual vertices

//Implemented by Pedro Company, 2017
{
	Input();

	//Reorder to place longer lines first
	ResetMerits();
	ReorderLines();

	long NumRemovedTips= 0;

	//Reify careful tips
	//------------------
	if(MergeCarefulLineTips !=0 && maxDistance !=0){
		//Find neighbor tips by way of Delaunay triangulation
		Delaunay D;
		if(D.NeighborTips(MJ_main, Neighbor)){
			//Merge close tips
			NumRemovedTips+= MergeCarefulTips();
		}
		else{
			AfxMessageBox("Tips of careful lines not merged, as no valid triangulation was found");
		}
		Neighbor.clear();
	}

	//Emerge casual tips
	//------------------
	if(MergeCasualLineTips !=0){
		//Merge dangling triplets of tips
		if(ReMergeTriplets && maxDistanceTriplets !=0){
			NumRemovedTips+= mergeCasualTriplets();
		}

		//Merge dangling isolated tips
		if(ReMergeDangling && maxDistanceDangling !=0){
			NumRemovedTips+= mergeCasualDanglingLines();
		}
	}

	for(long i=0; i<(long)MJ_main.VertexCount; i++){
		Vertices.push_back(MJ_main.Junction[i]);
	}
	for(long i=0; i<(long)MJ_main.EdgeCount; i++){
		HeadEdges.push_back(MJ_main.EdgeHead[i]);
		TailEdges.push_back(MJ_main.EdgeTail[i]);
	}

	return NumRemovedTips;
}
