//+-----------------------------------------------------------------------------+
//| Copyleft (C) 2016 REGEO (http://www.regeo.uji.es/)
//|
//| This program is free software
//| You can redistribute it and/or modify it under the terms of the GNU General Public License,
//| as published by the Free Software Foundation.
//| 
//| This program is distributed in the hope that it will be useful,
//| but WITHOUT ANY WARRANTY; without even the implied warranty of
//| MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//+----------------------------------------------------------------------------*/

#if !defined JUNCTIONS_DEFINED
	#define JUNCTIONS_DEFINED

	class CModelJunctions
	{
		public:
			int VertexCount;
			std::vector <POINT2D> Junction;

			int EdgeCount;
			std::vector <long> EdgeHead;
			std::vector <long> EdgeTail;

			void SetPOINT2D(long Position, POINT2D const &POINT2D);
			double GetLength2D(long IndexEdge);
			void RemovePOINT2D(long Position, bool updateEdges);
			void RemoveP(long Position);
			std::vector<long> GetAllEdgesInVertex(long iVertex);
			long GetSizeEdgesInVertex(long iPoint);

		CModelJunctions();
		CModelJunctions(const CModelJunctions& MJ);
		~CModelJunctions();
	};

class Junctions2D
	{
		private:
			//Input data
			//----------
			CModelJunctions MJ_main;  //Full model
			
			void Input();

			class Delaunay
			{
				private:
					struct EDGE{int p1, p2;};
					struct TRIANGLE{int p1, p2, p3;};

					bool ReorderPoints(std::vector<POINT2D> &pp,
									   std::vector<long> &NumTip);
					bool CircumCircle(double xp, double yp, double x1, double y1,
									  double x2, double y2, double x3, double y3,
									  double &xc, double &yc, double &);
					void Triangulate(std::vector<POINT2D> point,
									 std::vector<TRIANGLE> &v,
									 int &ntri);

					void addNewTip(long n, std::vector<long> &v);
				public:
					bool NeighborTips(CModelJunctions MJ_main,
									  std::vector<std::vector<long>> &nt);
			};

			std::vector<long> PriorLine;
			std::vector<double> LengthEdge;

			std::vector<std::vector<long>> Neighbor;
			std::vector<std::vector<long>> semiLines;

			struct SLX{
				std::vector<long> sl;
				std::vector<long> lineX;
				std::vector<POINT2D> pointX;
			};
			std::vector<SLX> sLines;

			std::vector<long> valence;
			std::vector<std::vector<long>> dsLines;

			long MergeCarefulLineTips;
			double maxDistance;
			double maxRotation;
			double ReificationMerit;
			long MergeCasualLineTips;
			double CasualInTol;
			double CasualOutTol;

			long ReMergeTriplets;
			double maxDistanceTriplets;
			double EmergentTripletsMerit;
			long ReMergeDangling;
			double maxDistanceDangling;
			double EmergenceDanglingDeMerit;

			void ResetMerits();
			void ReorderLines();

			void GetAllSemiLines();
			double AvgLength(long numTip1, long numTip2);
			long GetFarTip(long CloseTip, long &line);
			double GetIntersection(double x1, double y1, double x2, double y2,
								   double x3, double y3, double x4, double y4,
								   double maxDist,
								   double &x, double &y);
			double MeritRotation(long NumTip, POINT2D newPoint);
			double MeritTips(long NumTip1, long NumTip2,
							  POINT2D &junction, bool &swapTips);
			void DeleteNeighbor(long numNeighbor);
			void MergeTips(long NumTip1, long NumTip2,
						   POINT2D junction, double meritT);
			long MergeCarefulTips();

			void GetTripletSemiLines();
			void RemoveDanglingSemiLine(long num);
			bool SemiLinesIntersect(double x1, double y1, double x11, double y11,
									double x2, double y2, double x22, double y22,
									double &ta, double &tb, POINT2D &px);
			void GetTripletsIntersections();
			void RenumberTripletTips(long v1, long v2, long v3);
			void MergeTriplet(long numDangling, long numNeighbor,
							  POINT2D center, double merit,
							  long line2, long line3);
			long GetTriplets();
			long GetPoorTriplets();
			long mergeCasualTriplets();

			void GetDanglingSemiLines();
			double alwDangling(POINT2D p1, POINT2D p11, long junction);
			double meritCasualDangling(POINT2D p1, POINT2D p11, double length,
									   long junction);
			long mergeCasualDanglingLines();

	public:
			Junctions2D(long i_MergeCarefulLineTips,
						double i_maxDistance,
						double i_maxRotation,
						double i_ReificationMerit,
						long i_MergeCasualLineTips,
						double CasualInTol,
						double CasualOutTol,
						long i_ReMergeTriplets,
						double i_maxDistanceTriplets,
						double i_EmergentTripletsMerit,
						long i_ReMergeDangling,
						double i_maxDistanceDangling,
						double i_EmergenceDanglingMerit);
			
			Junctions2D::~Junctions2D();

			long GetJunctions(std::vector<POINT2D> &Junctions, std::vector<long> &HeadEdges, std::vector<long> &TailEdges);
	};
#endif // JUNCTIONS_DEFINED