// Main.cpp
//Main flow to detect vertices
//from a graph-like representation of a polyhedrical shape

//Based on:
//    Company P., Varley P.A.C., Plumed R. and Camba J.D.
//    "Detection of vertices in sketched drawing of polyhedral shapes"
//    CIARP: Iberoamerican Congress on Pattern Recognition (2018)
//    Ruben Vera-Rodriguez, Sergio Velastin, Manuel Montes-y-Gomez(Program Chairs)
//

#include "stdafx.h"
#include <vector>

#include "Tools_Vector.h"
#include "Tools_Geometry.h"
#include "DataRefine2D.h"
#include "CueJunctions.h"

int _tmain(int argc, _TCHAR* argv[])
{
	//Assign vectors
	std::vector<POINT2D> Vertices;
	std::vector<long> HeadEdges;
	std::vector<long> TailEdges;
	
	//Detect vertices
	//----------------
	CDataRefine2D R2D;

	Junctions2D J2D(R2D.MergeCarefulLineTips,
					R2D.MergeMaxDistVertices,
					R2D.MergeMaxVerticesRotation,
					R2D.ReificationMerit,
					R2D.MergeCasualLineTips,
					R2D.ReMergeCasualInTol,
					R2D.ReMergeCasualOutTol,
					R2D.ReMergeVerticesTriplets,
					R2D.ReMergeMaxDistTriplets,
					R2D.EmergentTripletsMerit,
					R2D.ReMergeVerticesDangling,
					R2D.ReMergeMaxDistDangling,
					R2D.EmergenceDanglingDeMerit);

		long NumRemovedTips= J2D.GetJunctions(Vertices, HeadEdges, TailEdges);

	//Print output
	//------------


	printf("Number of junctions= %3d \n\n", Vertices.size());

	for(long Num= 0; Num<(long)Vertices.size(); Num++){
		printf("Junction %d: %3f, %3f  merit: %3f \n", Num, Vertices[Num].x, Vertices[Num].y, Vertices[Num].merit);
	}

	printf("\n Number of edges= %3d \n\n", HeadEdges.size());

	for(long iedg= 0; iedg<(long)HeadEdges.size(); iedg++){
		printf("Edge: %d, from junction %d to %d \n", iedg, HeadEdges[iedg], TailEdges[iedg]);
	}

	printf("\n");

	system("PAUSE");
	return 0;
}


